#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <libxml/parser.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <asm/types.h>
#include <linux/videodev2.h>

#include "config.h"

#define MODULE_INPUT
#include "module.h"
#include "xmlhelp.h"
#include "grab.h"
#include "unpalette.h"
#include "image.h"
#include "log.h"

#define MODNAME "input_v4l2"

#define V4L2_CONTROL_COUNT		18
#define V4L2_PRIVATE_CONTROL_COUNT	32

char *name = MODNAME;
char *version = VERSION;
char *deps[] = { NULL };

const char *control_node = "control";

static int camdev_size_def(xmlNodePtr);
static int camdev_size_set(int, int, int, char *);

struct v4l2_camdev_control {
	char *name;
	int valid;
	__u32 id;
	__s32 value;
};

struct v4l2_camdev
{
	int fd;
	struct v4l2_capability cap;
	struct v4l2_format format;
	struct v4l2_queryctrl control_brightness;
	int autobrightness;
	unsigned int buffer_length;
	unsigned char *buffer;
	struct v4l2_camdev_control controls[V4L2_CONTROL_COUNT];
	struct v4l2_camdev_control private_controls[V4L2_PRIVATE_CONTROL_COUNT];
};

struct v4l2_camdev_control control_definitions[] = {
	{ "brightness", 0, V4L2_CID_BRIGHTNESS, 0 },
	{ "contrast", 0, V4L2_CID_CONTRAST, 0 },
	{ "saturation", 0, V4L2_CID_SATURATION, 0 },
	{ "hue", 0, V4L2_CID_HUE, 0 },
	{ "blacklevel", 0, V4L2_CID_BLACK_LEVEL, 0 },
	{ "awb", 0, V4L2_CID_AUTO_WHITE_BALANCE, 0 },
	{ "dowb", 0, V4L2_CID_DO_WHITE_BALANCE, 0 },
	{ "redbalance", 0, V4L2_CID_RED_BALANCE, 0 },
	{ "bluebalance", 0, V4L2_CID_BLUE_BALANCE, 0 },
	{ "gamma", 0, V4L2_CID_GAMMA, 0 },
	{ "whiteness", 0, V4L2_CID_WHITENESS, 0 },
	{ "exposure", 0, V4L2_CID_EXPOSURE, 0 },
	{ "agc", 0, V4L2_CID_AUTOGAIN, 0 },
	{ "gain", 0, V4L2_CID_GAIN, 0 },
	{ "hflip", 0, V4L2_CID_HFLIP, 0 },
	{ "vflip", 0, V4L2_CID_VFLIP, 0 },
	{ "hcenter", 0, V4L2_CID_HCENTER, 0 },
	{ "vcenter", 0, V4L2_CID_VCENTER, 0 },
};

static int v4l2_to_cs(__u32 v4l2_pixelformat) {
	switch (v4l2_pixelformat) {
	case V4L2_PIX_FMT_RGB24:
		return CAMSOURCE_PALETTE_RGB24;
	case V4L2_PIX_FMT_BGR24:
		return CAMSOURCE_PALETTE_BGR24;	
	case V4L2_PIX_FMT_RGB32:
		return CAMSOURCE_PALETTE_RGBA32;
	case V4L2_PIX_FMT_BGR32:
		return CAMSOURCE_PALETTE_BGRA32;
	case V4L2_PIX_FMT_YUYV:
		return CAMSOURCE_PALETTE_YUYV;
	case V4L2_PIX_FMT_YUV422P:
		return CAMSOURCE_PALETTE_YUV422;
	case V4L2_PIX_FMT_RGB565:
		return CAMSOURCE_PALETTE_RGB565;
	case V4L2_PIX_FMT_RGB555:
		return CAMSOURCE_PALETTE_RGB555;
	case V4L2_PIX_FMT_YUV420:
		return CAMSOURCE_PALETTE_YUV420;
	case V4L2_PIX_FMT_GREY:
		return CAMSOURCE_PALETTE_GREY;
	default:
		return CAMSOURCE_PALETTE_NONE;
	}
}

static __u32 cs_to_v4l2(int cs_palette) {
	switch (cs_palette) {
	case CAMSOURCE_PALETTE_RGB24:
		return V4L2_PIX_FMT_RGB24;
	case CAMSOURCE_PALETTE_BGR24:
		return V4L2_PIX_FMT_BGR24;
	case CAMSOURCE_PALETTE_RGBA32:
		return V4L2_PIX_FMT_RGB32;
	case CAMSOURCE_PALETTE_BGRA32:
		return V4L2_PIX_FMT_BGR32;
	case CAMSOURCE_PALETTE_YUYV:
		return V4L2_PIX_FMT_YUYV;
	case CAMSOURCE_PALETTE_YUV422:
		return V4L2_PIX_FMT_YUV422P;
	case CAMSOURCE_PALETTE_RGB565:
		return V4L2_PIX_FMT_RGB565;
	case CAMSOURCE_PALETTE_RGB555:
		return V4L2_PIX_FMT_RGB555;
	case CAMSOURCE_PALETTE_YUV420:
		return V4L2_PIX_FMT_YUV420;
	case CAMSOURCE_PALETTE_GREY:
		return V4L2_PIX_FMT_GREY;
	default:
		return V4L2_PIX_FMT_GREY;
	}
}

int
opendev(xmlNodePtr node, struct grab_camdev *gcamdev)
{
	struct v4l2_camdev newcamdev;
	int ret;
	int i;

	char *path;
	int x, y, fps;
	int input;
	int norm;
	int autobrightness;
	char *dataformat;
	char *usedataformat;

	struct palette *pal, *usepal;

	struct v4l2_cropcap cropcap;
	struct v4l2_crop crop;
	v4l2_std_id std;
	struct v4l2_control ctrl;
	struct v4l2_streamparm streamparm;
	struct v4l2_requestbuffers reqbufs;
	struct v4l2_buffer buf;
	
	memset(&newcamdev, 0, sizeof(newcamdev));

	memcpy(&newcamdev.controls, &control_definitions,
	       V4L2_CONTROL_COUNT * sizeof(struct v4l2_camdev_control));
	
	path = "/dev/video0";
	x = y = fps = 0;
	input = norm = autobrightness = -1;
	dataformat = usedataformat = NULL;

	if (node)
	{
		for (node = node->xml_children; node; node = node->next)
		{
			if (xml_isnode(node, "path"))
				path = xml_getcontent_def(node, path);
			else if (xml_isnode(node, "width"))
				x = camdev_size_def(node);
			else if (xml_isnode(node, "height"))
				y = camdev_size_def(node);
			else if (xml_isnode(node, "fps"))
				fps = xml_atoi(node, 0);
			else if (xml_isnode(node, "input"))
				input = xml_atoi(node, -1);
			else if (xml_isnode(node, "autobrightness"))
				autobrightness = xml_atoi(node, 0);
			else if (xml_isnode(node, "dataformat"))
				dataformat = xml_getcontent_def(node, NULL);
			else if (xml_isnode(node, "usedataformat"))
				usedataformat =
					xml_getcontent_def(node, NULL);
			else if (xml_isnode(node, "norm")) {
				char *s = xml_getcontent_def(node, "");
				if (!strcasecmp(s, "pal")) {
					norm = 1;
					std = V4L2_STD_PAL;
				} else if (!strcasecmp(s, "ntsc")) {
					norm = 1;
					std = V4L2_STD_NTSC;
				} else if (!strcasecmp(s, "secam")) {
					norm = 1;
					std = V4L2_STD_SECAM;
				} else if (!strcasecmp(s, "auto")) {
					norm = 1;
					std = V4L2_STD_ALL;
				} else {
					log_log(MODNAME, "Invalid video norm \"%s\" "
						"specified, ignoring\n",
						s);
				}
			} else if (xml_isnnode(node, control_node,
					       strlen(control_node))) {
				unsigned int off = strlen(control_node);
				char *istr = (char *)((char *)node->name + off);
				int index = atoi(istr);
				int value = xml_atoi(node, 0);

				if ((index >= 0)
				    && (index < V4L2_PRIVATE_CONTROL_COUNT)) {
					newcamdev.private_controls[index].valid = 1;
					newcamdev.private_controls[index].id =
						V4L2_CID_PRIVATE_BASE + index;
					newcamdev.private_controls[index].value = value;
				} else {
					log_log(MODNAME, "Invalid private control "
						"\"%s\" specified, ignoring\n",
						node->name);
				}
			}
			for (i = 0; i < V4L2_CONTROL_COUNT; i++) {
				if (xml_isnode(node, newcamdev.controls[i].name)) {
					int value = xml_atoi(node, 0);
					newcamdev.controls[i].valid = 1;
					newcamdev.controls[i].value = value;
				}
			}
		}
	}
	
	newcamdev.fd = open(path, O_RDWR);
	if (newcamdev.fd < 0)
	{
		log_log(MODNAME, "Unable to open %s (%s)\n", path, strerror(errno));
		return -1;
	}
	
	ret = ioctl(newcamdev.fd, VIDIOC_QUERYCAP, &newcamdev.cap);
	if (ret != 0)
	{
		log_log(MODNAME, "The device is not a V4L2-device: "
			"ioctl \"querycap\" failed: %s\n",
			strerror(errno));
closenerr:
		close(newcamdev.fd);
		return -1;
	}
	if (!(newcamdev.cap.capabilities & V4L2_CAP_VIDEO_CAPTURE))
	{
		log_log(MODNAME, "Video device doesn't support capturing\n");
		goto closenerr;
	}
	
	if (input >= 0)
	{
		ret = ioctl(newcamdev.fd, VIDIOC_S_INPUT, &input);
		if (ret)
			log_log(MODNAME, "ioctl \"set input\" failed, "
				"continuing anyway: %s\n",
				strerror(errno));
	}
	if (norm >= 0) {
		ret = ioctl(newcamdev.fd, VIDIOC_S_STD, &std);
		if (ret)
			log_log(MODNAME, "ioctl \"set standard\" failed, "
				"continuing anyway: %s\n",
				strerror(errno));
	}	

	newcamdev.control_brightness.id = V4L2_CID_BRIGHTNESS;
	ret = ioctl(newcamdev.fd, VIDIOC_QUERYCTRL,
		    &newcamdev.control_brightness);
	if (ret) {
		newcamdev.autobrightness = -1;
		log_log(MODNAME, "ioctl \"query control\""
			"failed, continuing anyway: %s\n", strerror(errno));
	} else {
		newcamdev.autobrightness = autobrightness;
	}

	memset(&cropcap, 0, sizeof(cropcap));
	cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	ret = ioctl(newcamdev.fd, VIDIOC_CROPCAP, &cropcap);
	if (ret) {
		log_log(MODNAME, "ioctl \"get crop capabilities\" failed, "
			"cropping isn't probably supported: %s\n",
			strerror(errno));
	} else {
		x = camdev_size_set(x, 0, cropcap.defrect.width, "width");
		y = camdev_size_set(y, 0, cropcap.defrect.height, "height");
		if (x <= 0 || y <= 0)
			goto closenerr;

		memset(&crop, 0, sizeof(crop));
		crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		crop.c.left = cropcap.defrect.left;
		crop.c.top = cropcap.defrect.top;
		crop.c.width = cropcap.defrect.width;
		crop.c.height = cropcap.defrect.height;

		ret = ioctl(newcamdev.fd, VIDIOC_S_CROP, &crop);
		if (ret) {
			log_log(MODNAME, "ioctl \"set crop\" failed, continuing anyway: %s\n",
				strerror(errno));
		}
	}
	
	memset(&newcamdev.format, 0, sizeof(newcamdev.format));
	newcamdev.format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	ret = ioctl(newcamdev.fd, VIDIOC_G_FMT, &newcamdev.format);
	if (ret)
	{
		log_log(MODNAME, "ioctl \"get format\" failed: %s\n", strerror(errno));
		goto closenerr;
	}

	if (dataformat == NULL) {
autopal:
		for (pal = palettes; pal->val >= 0; pal++) {
			newcamdev.format.fmt.pix.width = x;
			newcamdev.format.fmt.pix.height = y;
			newcamdev.format.fmt.pix.field = V4L2_FIELD_INTERLACED;
			newcamdev.format.fmt.pix.pixelformat =
				cs_to_v4l2(pal->val);

			ioctl(newcamdev.fd, VIDIOC_S_FMT, &newcamdev.format);
			ioctl(newcamdev.fd, VIDIOC_G_FMT, &newcamdev.format);
			if (v4l2_to_cs(newcamdev.format.fmt.pix.pixelformat)
			    == pal->val)
				goto palfound;	/* break */
		}
		log_log(MODNAME, "No common supported palette found\n");
		goto closenerr;
	} else {
		for (pal = palettes; pal->val >= 0; pal++) {
			if (!strcasecmp(pal->id, dataformat))
				goto trypal;
		}
		log_log(MODNAME, "Invalid data format \"%s\" specified\n",
			dataformat);
		dataformat = NULL;
		goto autopal;

trypal:
		newcamdev.format.fmt.pix.width = x;
		newcamdev.format.fmt.pix.height = y;
		newcamdev.format.fmt.pix.field = V4L2_FIELD_INTERLACED;
		newcamdev.format.fmt.pix.pixelformat =
			cs_to_v4l2(pal->val);

		ioctl(newcamdev.fd, VIDIOC_S_FMT, &newcamdev.format);
		ioctl(newcamdev.fd, VIDIOC_G_FMT, &newcamdev.format);
		if (v4l2_to_cs(newcamdev.format.fmt.pix.pixelformat)
		    == pal->val)
			goto palfound;	/* break */

		log_log(MODNAME, "Device doesn't support data format \"%s\"\n",
			dataformat);
		dataformat = NULL;
		goto autopal;
	}
	
palfound:
	if ((dataformat != NULL) && (usedataformat != NULL)) {
		for (usepal = palettes; usepal->val >= 0; usepal++) {
			if (!strcasecmp(usepal->id, usedataformat))
				goto tryusepal;
		}
		log_log(MODNAME, "Invalid data format \"%s\" specified for use "
			"instead of the driver's format, ignoring\n",
			usedataformat);
		usepal = pal;
		goto over;

tryusepal:
		if (pal->bpp != usepal->bpp) {
			log_log(MODNAME, "The depth of actually used data format \"%s\" "
				"doesn't match the depth of the selected "
				"data format \"%s\", ignoring\n",
				usedataformat, dataformat);
			usepal = pal;
		}
	} else {
		usepal = pal;
	}

over:
	memset(&ctrl, 0, sizeof(struct v4l2_control));
	for (i = 0; i < V4L2_CONTROL_COUNT; i++) {
		if (!newcamdev.controls[i].valid)
			continue;

		ctrl.id = newcamdev.controls[i].id;
		ctrl.value = newcamdev.controls[i].value;

		ret = ioctl(newcamdev.fd, VIDIOC_S_CTRL, &ctrl);
		if (ret) {
			log_log(MODNAME, "ioctl \"set control\" for \"%s\" failed, "
				"continuing anyway: %s\n", newcamdev.controls[i].name,
				strerror(errno));
		}
	}

	for (i = 0; i < V4L2_PRIVATE_CONTROL_COUNT; i++) {
		if (!newcamdev.private_controls[i].valid)
			continue;

		ctrl.id = newcamdev.private_controls[i].id;
		ctrl.value = newcamdev.private_controls[i].value;

		ret = ioctl(newcamdev.fd, VIDIOC_S_CTRL, &ctrl);
		if (ret) {
			log_log(MODNAME, "ioctl \"set control\" for private control %d"
				"failed, continuing anyway: %s\n",
				i, strerror(errno));
		}
	}

	if (newcamdev.cap.capabilities & V4L2_CAP_STREAMING) {
		if (fps > 0) {
			memset(&streamparm, 0, sizeof(streamparm));
			streamparm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			ret = ioctl(newcamdev.fd, VIDIOC_G_PARM, &streamparm);
			if (!ret && (streamparm.parm.capture.capability
				     & V4L2_CAP_TIMEPERFRAME)) {
				memset(&streamparm, 0, sizeof(streamparm));
				streamparm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
				streamparm.parm.capture.
					timeperframe.numerator = 1;
				streamparm.parm.capture.
					timeperframe.denominator = fps;
				ret = ioctl(newcamdev.fd, VIDIOC_S_PARM,
					    &streamparm);
				if (ret) {
					log_log(MODNAME, "ioctl \"set streaming "
						"parameters\" failed, "
						"continuing anyway: %s\n",
						strerror(errno));
				}
			}
		}

		memset(&reqbufs, 0, sizeof(reqbufs));
		reqbufs.count = 1;
		reqbufs.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		reqbufs.memory = V4L2_MEMORY_MMAP;
		ret = ioctl(newcamdev.fd, VIDIOC_REQBUFS, &reqbufs);
		if (ret) {
			log_log(MODNAME, "ioctl \"request buffers\" failed: %s\n",
				strerror(errno));
			goto closenerr;
		}
		if (reqbufs.count < 1) {
			log_log(MODNAME, "Not enough memory to allocate buffers.");
			goto closenerr;
		}

		memset(&buf, 0, sizeof(buf));
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = 0;
		ret = ioctl(newcamdev.fd, VIDIOC_QUERYBUF, &buf);
		if (ret) {
			log_log(MODNAME, "ioctl \"query buffer\" failed: %s\n",
			       strerror	(errno));
			goto closenerr;
		}
		newcamdev.buffer_length = buf.length;
		newcamdev.buffer = mmap(NULL, buf.length, PROT_READ
					| PROT_WRITE, MAP_SHARED,
					newcamdev.fd, buf.m.offset);
		if (newcamdev.buffer == MAP_FAILED) {
			log_log(MODNAME, "mmap failed: %s\n", strerror(errno));
			goto closenerr;
		}
		ret = ioctl(newcamdev.fd, VIDIOC_STREAMON, &buf);
		if (ret) {
			log_log(MODNAME, "ioctl \"set streaming on\" failed: %s\n",
				strerror(errno));
			goto closenerr;
		}
	} else if (newcamdev.cap.capabilities & V4L2_CAP_READWRITE) {
		newcamdev.buffer_length = newcamdev.format.fmt.pix.sizeimage;
		newcamdev.buffer = malloc(newcamdev.buffer_length);
	} else {
		log_log(MODNAME, "Device doesn't support streaming "
			"or direct reading.\n");
		goto closenerr;
	}


	gcamdev->name = strdup(path);
	gcamdev->pal = usepal;
	gcamdev->x = newcamdev.format.fmt.pix.width;
	gcamdev->y = newcamdev.format.fmt.pix.height;

	gcamdev->custom = malloc(sizeof(newcamdev));
	memcpy(gcamdev->custom, &newcamdev, sizeof(newcamdev));
	
	return 0;
}

unsigned char *
input(struct grab_camdev *gcamdev)
{
	int ret;
	struct v4l2_camdev *camdev;
	struct v4l2_buffer buf;

	camdev = gcamdev->custom;

	if (camdev->cap.capabilities & V4L2_CAP_STREAMING) { // mmap
		memset(&buf, 0, sizeof(buf));
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = 0;
		ret = ioctl(camdev->fd, VIDIOC_QBUF, &buf);
		if (ret) {
			if (errno != EINVAL) {
				log_log(MODNAME, "ioctl \"queue buffer\" failed: %s\n",
					strerror(errno));
				return NULL;
			}
		}
retry:
		ret = ioctl(camdev->fd, VIDIOC_DQBUF, &buf);
		if (ret) {
			if ((errno == EIO) || (errno == EAGAIN)) {
				/* probably signal loss */
				if (buf.flags & V4L2_BUF_FLAG_QUEUED) {
					goto retry;
				}
			} else {
				log_log(MODNAME, "ioctl \"dequeue buffer\" failed: %s\n",
					strerror(errno));
				return NULL;
			}
		}
	} else { // read
		do {
			ret = read(camdev->fd, camdev->buffer,
				   camdev->buffer_length);
		} while ((ret < 0) && (errno == EINTR));

		if (ret < 0) {
			log_log(MODNAME, "Error while reading from device '%s': %s\n",
				gcamdev->name, strerror(errno));
			return NULL;
		}
		if (ret == 0) {
			log_log(MODNAME, "EOF while reading from device '%s'\n",
				gcamdev->name);
			return NULL;
		}
		if (ret < camdev->buffer_length)
			log_log(MODNAME, "Short read while reading from device '%s' "
				"(%i < %i), continuing anyway\n",
				gcamdev->name, ret, camdev->buffer_length);
	}
	
	return camdev->buffer;
}

static
int
camdev_size_def(xmlNodePtr node)
{
	char *s;
	
	s = xml_getcontent_def(node, "max");
	if (!strcmp(s, "max") || !strcmp(s, "maximum")
	    || !strcmp(s, "default"))
		return 0;
	else if (!strcmp(s, "min") || !strcmp(s, "minimum"))
		return -1;
	else
		return xml_atoi(node, 0);
}

static
int
camdev_size_set(int val, int min, int max, char *s)
{
	if (val == 0)
		return max;
	if (val == -1)
		return min;
	if (val < min)
	{
		log_log(MODNAME, "Invalid grabbing %s according to driver (%i < %i)\n",
			s, val, min);
		return 0;
	}
	if (val > max)
	{
		log_log(MODNAME, "Invalid grabbing %s according to driver (%i > %i)\n",
			s, val, max);
		return 0;
	}
	return val;
}

static __s32 calculate_change(__s32 value, __s32 step,
			      __s32 minimum, __s32 maximum,
			      __u32 current, __u32 target) {
	__u64 min = 0;
	__u64 max = maximum;
	__u64 val = value;
	__u64 newval;
	__s32 change;

	if ((current >= (target - 1)) && (current <= (target + 1))) {
		// log_log(MODNAME, "No change needed\n");
		return 0;
	}

	if (minimum < 0) {
		max += (-minimum);
		val += (-minimum);
	}

	change = target - current;

	// log_log(MODNAME, "Change: %d, Value: %d\n", change, value);
	if ((val == 0) && (max > 0)) {
		val = 1;
	}

	double div = (100 * (65535 / (double)max));
	change = (__u64)((double)val * ((double)change / div) * 3);

	// log_log(MODNAME, "Change: %d, Div: %f\n", change, div);

	newval = val + change;

	if (newval > max) {
		change = max - val;
	} else if (newval < min) {
		change = -(val - min);
	}

	return (__s32)change;
}

void
postprocess(struct grab_camdev *gcamdev, struct image *img)
{
	struct v4l2_camdev *camdev;
	struct v4l2_control ctrl;
	float brightness;
	int ret;

	camdev = gcamdev->custom;

	if (camdev->autobrightness <= 0)
		return;

	brightness = image_brightness(img);
	// log_log(MODNAME, "Current Image Brightness = %2.1f\n", brightness);

	ctrl.id = V4L2_CID_BRIGHTNESS;

	ret = ioctl(camdev->fd, VIDIOC_G_CTRL, &ctrl);
	if (ret) {
		log_log(MODNAME, "ioctl \"get control\" failed: %s\n",
			strerror(errno));
	}

	__s32 change = calculate_change(ctrl.value,
					camdev->control_brightness.step,
					camdev->control_brightness.minimum,
					camdev->control_brightness.maximum,
					brightness, camdev->autobrightness);

	// log_log(MODNAME, "Current value = %d, Calculated change = %d\n", ctrl.value, change);

	ctrl.id = V4L2_CID_BRIGHTNESS;
	ctrl.value += change;

	ret = ioctl(camdev->fd, VIDIOC_S_CTRL, &ctrl);
	if (ret) {
		log_log(MODNAME, "ioctl \"set control\" failed: %s\n",
			strerror(errno));
	}
}

static void enumerate_menu(int fd, struct v4l2_queryctrl *queryctrl)
{
	struct v4l2_querymenu querymenu;

        memset(&querymenu, 0, sizeof (querymenu));
        querymenu.id = queryctrl->id;

        for (querymenu.index = queryctrl->minimum;
             querymenu.index <= queryctrl->maximum;
	     querymenu.index++) {
                if (ioctl(fd, VIDIOC_QUERYMENU, &querymenu) == 0) {
                        log_log(MODNAME, "    %s\n", querymenu.name);
                } else {
			log_log(MODNAME, "ioctl(VIDIOC_QUERYMENU) (query menu) "
				"failed: %s\n", strerror(errno));
			break;
                }
        }
}

static void enumerate_control(int fd, struct v4l2_queryctrl *queryctrl)
{
	char *type = "unknown";

	if (queryctrl->type == V4L2_CTRL_TYPE_INTEGER) {
		type = "integer";
	} else if (queryctrl->type == V4L2_CTRL_TYPE_BOOLEAN) {
		type = "boolean";
	} else if (queryctrl->type == V4L2_CTRL_TYPE_BUTTON) {
		type = "button ";
	}

	log_log(MODNAME, "    Type: %s Minimum: %05d Maximum: %05d "
		"Default: %05d Step: %03d\n",
		type, queryctrl->minimum, queryctrl->maximum,
		queryctrl->default_value, queryctrl->step);
}

void
capdump(xmlNodePtr node, struct grab_camdev *gcamdev)
{
	int fd, ret;

	char *path;
	int input;

	struct v4l2_capability cap;
	struct v4l2_input input_info;
	struct v4l2_format format;
	struct palette *pal;
	struct v4l2_queryctrl queryctrl;

	path = "/dev/video0";
	input = -1;

	if (node)
	{
		for (node = node->xml_children; node; node = node->next)
		{
			if (xml_isnode(node, "path"))
				path = xml_getcontent_def(node, path);
			else if (xml_isnode(node, "input"))
				input = xml_atoi(node, -1);
		}
	}

	fd = open(path, O_RDWR);
	if (fd < 0)
	{
		log_log(MODNAME, "Unable to open %s (%s)\n", path, strerror(errno));
		return;
	}

	ret = ioctl(fd, VIDIOC_QUERYCAP, &cap);
	if (ret < 0)
	{
		log_log(MODNAME, "ioctl(VIDIOC_QUERYCAP) (query video capabilites) failed: %s\n",
			strerror(errno));
		goto closenout;
	}
	
	log_log(MODNAME, "Capability info for %s:\n", path);
	log_log(MODNAME, "  Name: %s\n", cap.driver);
	log_log(MODNAME, "    Card: %s\n", cap.card);
	log_log(MODNAME, "    Bus: %s\n", cap.bus_info);
	log_log(MODNAME, "    Can %scapture to memory\n",
		(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) ? "" : "not ");
	log_log(MODNAME, "    Overlay is %ssupported\n",
		(cap.capabilities & V4L2_CAP_VIDEO_OVERLAY) ? "" : "not ");
	log_log(MODNAME, "    Streaming is %ssupported\n",
		(cap.capabilities & V4L2_CAP_STREAMING) ? "" : "not ");
	log_log(MODNAME, "    Direct read/write is %ssupported\n",
		(cap.capabilities & V4L2_CAP_READWRITE) ? "" : "not ");
	log_log(MODNAME, "    %s a tuner\n",
		(cap.capabilities & V4L2_CAP_TUNER) ? "Has" : "Doesn't have");
	log_log(MODNAME, "    Can%s receive teletext\n",
		(cap.capabilities & V4L2_CAP_VBI_CAPTURE) ? "" : "not");
	log_log(MODNAME, "    Can%s output video\n",
		(cap.capabilities & V4L2_CAP_VIDEO_OUTPUT) ? "" : "not");
	log_log(MODNAME, "    Can%s output teletext\n",
		(cap.capabilities & V4L2_CAP_VBI_OUTPUT) ? "" : "not");

	if (input >= 0)
	{
		ret = ioctl(fd, VIDIOC_S_INPUT, &input);
		if (ret)
			log_log(MODNAME, "ioctl \"set input\" failed, "
				"continuing anyway: %s\n",
				strerror(errno));
	}

	ret = ioctl(fd, VIDIOC_G_INPUT, &input);
	if (ret) {
		log_log(MODNAME, "ioctl \"get input\" failed: %s\n",
			strerror(errno));
		goto closenout;
	}

	input_info.index = input;
	ret = ioctl(fd, VIDIOC_ENUMINPUT, &input_info);
	if (ret) {
		log_log(MODNAME, "ioctl \"enum input\" failed: %s\n",
			strerror(errno));
		goto closenout;
	}

	log_log(MODNAME, "\n");
	log_log(MODNAME, "Input %d:\n", input);
	log_log(MODNAME, "  Name: %s\n", input_info.name);
	log_log(MODNAME, "    Type: %s\n",
		(input_info.type == V4L2_INPUT_TYPE_TUNER)
		? "tuner" : "camera");
	log_log(MODNAME, "    Power: %s\n",
		(input_info.status & V4L2_IN_ST_NO_POWER) ? "off" : "on");
	log_log(MODNAME, "    Signal %sdetected\n",
		(input_info.status & V4L2_IN_ST_NO_SIGNAL) ? "not " : "");
	log_log(MODNAME, "    Color %sdetected\n",
		(input_info.status & V4L2_IN_ST_NO_COLOR) ? "not " : "");

	memset(&format, 0, sizeof(format));
	format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	ret = ioctl(fd, VIDIOC_G_FMT, &format);
	if (ret != 0)
	{
		log_log(MODNAME, "ioctl(VIDIOC_G_FMT) (get format) failed: %s\n",
			strerror(errno));
		goto closenout;
	}
	
	log_log(MODNAME, "\n");
	log_log(MODNAME, "Palette information:\n");
	for (pal = palettes; pal->val >= 0; pal++)
	{
		if (pal->val == v4l2_to_cs(format.fmt.pix.pixelformat))
		{
			log_log(MODNAME, "  Currently active palette: "
				"%s with depth %u\n",
				pal->name, pal->depth);
			goto palfound;
		}
	}
	log_log(MODNAME, "  Currently active palette: not found/supported? "
	       "(value %08x)\n", format.fmt.pix.pixelformat);
	
palfound:
	log_log(MODNAME, "  Probing for supported palettes:\n");
	for (pal = palettes; pal->val >= 0; pal++)
	{
		format.fmt.pix.pixelformat = cs_to_v4l2(pal->val);
		ioctl(fd, VIDIOC_S_FMT, &format);
		ioctl(fd, VIDIOC_G_FMT, &format);
		if (v4l2_to_cs(format.fmt.pix.pixelformat) == pal->val)
			log_log(MODNAME, "    Palette \"%s\" supported: "
				"Yes, with depth %u\n",
				pal->name, pal->depth);
		else
			log_log(MODNAME, "    Palette \"%s\" supported: "
				"No\n", pal->name);
	}	
	log_log(MODNAME, "\n");
	log_log(MODNAME, "Enumerating image controls:\n");

	memset(&queryctrl, 0, sizeof(queryctrl));

	__u32 id;
	for (id = V4L2_CID_BASE; id < V4L2_CID_LASTP1; id++) {
		queryctrl.id = id;
		ret = ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl);
		if (ret == 0) {
			if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
				continue;

			log_log(MODNAME, "  Control: %s\n", queryctrl.name);

			if (queryctrl.type == V4L2_CTRL_TYPE_MENU) {
				enumerate_menu(fd, &queryctrl);
			} else {
				enumerate_control(fd, &queryctrl);
			}
		} else {
			if (errno == EINVAL)
				continue;

			log_log(MODNAME, "ioctl(VIDIOC_QUERYCTRL) (query control) "
				"failed: %s\n", strerror(errno));
			goto closenout;
		}
	}

	log_log(MODNAME, "\n");
	log_log(MODNAME, "  Enumerating private (driver-specific) image controls:\n");

	for (id = V4L2_CID_PRIVATE_BASE;; id++) {
		queryctrl.id = id;
		ret = ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl);
		if (ret == 0) {
			if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
				continue;

			log_log(MODNAME, "  Control %03d: %s\n",
			       queryctrl.id - V4L2_CID_PRIVATE_BASE,
			       queryctrl.name);

			if (queryctrl.type == V4L2_CTRL_TYPE_MENU) {
				enumerate_menu(fd, &queryctrl);
			} else {
				enumerate_control(fd, &queryctrl);
			}
		} else {
			if (errno == EINVAL)
				break;

			log_log(MODNAME, "ioctl(VIDIOC_QUERYCTRL) (query control) "
				"failed: %s\n", strerror(errno));
			goto closenout;
		}
	}

closenout:
	close(fd);
}
